import { generate } from '@babel/generator';
import { parse } from '@babel/parser';

const DEFINE_EMITS = "defineEmits";
const DEFINE_PROPS = "defineProps";
const WITH_DEFAULTS = "withDefaults";
const DEFINE_MODEL = "defineModel";
async function preTranspileScriptSetup(sfc, id) {
  if (!sfc.scriptSetup) {
    throw new Error("No script setup block found");
  }
  const context = await prepareContext(sfc, id);
  const resultBuilder = new context.utils.MagicString(sfc.scriptSetup.content);
  for (const node of context.ctx.ast) {
    if (node.type === "ExpressionStatement") {
      const processedTypeSyntax = processDefineProps(node.expression, context) || processDefineEmits(node.expression, context) || processWithDefaults(node.expression, context) || processDefineModel(node.expression, context);
      if (processedTypeSyntax !== void 0) {
        resultBuilder.overwrite(node.start, node.end, processedTypeSyntax);
      }
    }
    if (node.type === "VariableDeclaration" && !node.declare) {
      const total = node.declarations.length;
      for (let i = 0; i < total; i += 1) {
        const decl = node.declarations[i];
        if (!decl.init) {
          continue;
        }
        const processedTypeSyntax = processDefineProps(decl.init, context) || processDefineEmits(decl.init, context) || processWithDefaults(decl.init, context) || processDefineModel(decl.init, context);
        if (processedTypeSyntax !== void 0) {
          resultBuilder.overwrite(
            decl.init.start,
            decl.init.end,
            processedTypeSyntax
          );
        }
      }
    }
  }
  return {
    ...sfc.scriptSetup,
    content: resultBuilder.toString()
  };
}
function processDefineProps(node, context) {
  if (!isCallOf(node, DEFINE_PROPS)) {
    return;
  }
  const propsRuntimeDecl = node.arguments[0];
  if (!node.typeParameters) {
    return;
  }
  if (propsRuntimeDecl) {
    context.ctx.error(
      `${DEFINE_PROPS}() cannot accept both type and non-type arguments at the same time. Use one or the other.`,
      node
    );
  }
  const propsTypeDecl = node.typeParameters.params[0];
  if (!propsTypeDecl) {
    return;
  }
  context.ctx.propsTypeDecl = propsTypeDecl;
  const propsStr = context.utils.extractRuntimeProps(context.ctx);
  const propsAst = parse(`${DEFINE_PROPS}(${propsStr})`, {
    sourceType: "module",
    plugins: ["typescript"]
  });
  node.typeArguments = void 0;
  node.typeParameters = void 0;
  node.arguments = propsAst.program.body[0].expression.arguments;
  return generate(node).code;
}
function processDefineEmits(node, context) {
  if (!isCallOf(node, DEFINE_EMITS)) {
    return;
  }
  const emitsRuntimeDecl = node.arguments[0];
  if (!node.typeParameters) {
    return;
  }
  if (emitsRuntimeDecl) {
    context.ctx.error(
      `${DEFINE_EMITS}() cannot accept both type and non-type arguments at the same time. Use one or the other.`,
      node
    );
  }
  const emitsTypeDecl = node.typeParameters.params[0];
  if (!emitsTypeDecl) {
    return;
  }
  context.ctx.emitsTypeDecl = emitsTypeDecl;
  const emits = context.utils.extractRuntimeEmits(context.ctx);
  node.typeArguments = void 0;
  node.typeParameters = void 0;
  node.arguments[0] = {
    type: "ArrayExpression",
    elements: [...emits].map((emit) => ({
      type: "StringLiteral",
      value: emit
    }))
  };
  return generate(node).code;
}
function processWithDefaults(node, context) {
  if (!isCallOf(node, WITH_DEFAULTS)) {
    return;
  }
  context.ctx.propsRuntimeDefaults = node.arguments[1];
  const res = processDefineProps(node.arguments[0], context);
  if (!res) {
    context.ctx.error(
      `${WITH_DEFAULTS}' first argument must be a ${DEFINE_PROPS} call.`,
      node.arguments[0] || node
    );
  }
  if (!context.ctx.propsTypeDecl) {
    context.ctx.error(
      `${WITH_DEFAULTS} can only be used with type-based ${DEFINE_PROPS} declaration.`,
      node
    );
  }
  if (!context.ctx.propsRuntimeDefaults) {
    context.ctx.error(
      `The 2nd argument of ${WITH_DEFAULTS} is required.`,
      node
    );
  }
  return res;
}
function processDefineModel(node, context) {
  if (!isCallOf(node, DEFINE_MODEL)) {
    return;
  }
  const [modelNameDecl, modelRuntimeDecl] = getDefineModelRuntimeDecl(
    node,
    context
  );
  const modelTypeDecl = node.typeParameters?.params[0];
  if (!modelTypeDecl) {
    return;
  }
  const model = context.utils.inferRuntimeType(context.ctx, modelTypeDecl);
  if (!model || model.length === 0 || model[0] === "Unknown") {
    return;
  }
  const modelCodegenTypeDecl = model.length === 1 ? { type: "Identifier", name: model[0] } : {
    type: "ArrayExpression",
    elements: model.map(
      (name) => ({ type: "Identifier", name })
    )
  };
  const modelCodegenDecl = {
    type: "ObjectExpression",
    properties: [
      {
        type: "ObjectProperty",
        key: { type: "StringLiteral", value: "type" },
        value: modelCodegenTypeDecl,
        computed: false,
        shorthand: false
      }
    ]
  };
  if (modelRuntimeDecl) {
    modelCodegenDecl.properties.push({
      type: "SpreadElement",
      argument: modelRuntimeDecl
    });
  }
  node.typeArguments = void 0;
  node.typeParameters = void 0;
  node.arguments = modelNameDecl ? [modelNameDecl, modelCodegenDecl] : [modelCodegenDecl];
  return generate(node).code;
}
function getDefineModelRuntimeDecl(node, context) {
  const [arg0, arg1] = node.arguments;
  if (arg0 && arg0.type === "StringLiteral") {
    if (arg1 && arg1.type !== "ObjectExpression") {
      context.ctx.error(
        `${DEFINE_MODEL}()'s second argument must be an object.`,
        arg1
      );
    }
    return [arg0, arg1];
  }
  if (arg0 && arg0.type !== "ObjectExpression" && !(arg0.type === "Identifier" && arg0.name === "undefined")) {
    context.ctx.error(`Unexpected argument type for ${DEFINE_MODEL}().`, arg0);
  }
  return [void 0, arg0];
}
async function prepareContext({ script, scriptSetup }, id) {
  const { extractRuntimeProps, extractRuntimeEmits, inferRuntimeType, MagicString } = await import('vue/compiler-sfc');
  const helper = /* @__PURE__ */ new Set();
  const ast = parse(`${scriptSetup.content}
${script?.content}`, {
    sourceType: "module",
    plugins: ["tsx", "jsx"].includes(scriptSetup.lang) ? ["typescript", "jsx"] : ["typescript"]
  });
  const ctx = {
    filename: id,
    source: scriptSetup.content,
    ast: ast.program.body,
    error: (msg) => {
      throw new Error(`[vue-sfc-transformer] ${msg}`);
    },
    helper: (key) => {
      helper.add(key);
      return `_${key}`;
    },
    getString: (node) => {
      return scriptSetup.content.slice(node.start, node.end);
    },
    propsTypeDecl: void 0,
    propsRuntimeDefaults: void 0,
    propsDestructuredBindings: /* @__PURE__ */ Object.create(null),
    emitsTypeDecl: void 0,
    isCE: false,
    options: {}
  };
  return {
    ctx,
    utils: {
      MagicString,
      extractRuntimeProps,
      extractRuntimeEmits,
      inferRuntimeType
    }
  };
}
function isCallOf(node, test) {
  return !!(node && node.type === "CallExpression" && node.callee.type === "Identifier" && (typeof test === "string" ? node.callee.name === test : test(node.callee.name)));
}

function handleNode(node, addExpression) {
  if (!node) {
    return;
  }
  const search = (node2) => handleNode(node2, addExpression);
  switch (node.type) {
    case 0 /* ROOT */: {
      for (const child of node.children) {
        search(child);
      }
      return;
    }
    case 1 /* ELEMENT */: {
      const nodes = [...node.children, ...node.props];
      for (const child of nodes) {
        search(child);
      }
      return;
    }
    case 2 /* TEXT */: {
      return;
    }
    case 3 /* COMMENT */: {
      return;
    }
    case 4 /* SIMPLE_EXPRESSION */: {
      if (node.ast === null || node.ast === false) {
        return;
      }
      addExpression({ loc: node.loc, src: node.content });
      return;
    }
    case 5 /* INTERPOLATION */: {
      search(node.content);
      return;
    }
    case 6 /* ATTRIBUTE */: {
      search(node.value);
      return;
    }
    case 7 /* DIRECTIVE */: {
      const nodes = [
        node.exp,
        // node.arg,
        // node.forParseResult?.source,
        // node.forParseResult?.value,
        // node.forParseResult?.key,
        // node.forParseResult?.index,
        ...node.modifiers
      ].filter((item) => !!item);
      for (const child of nodes) {
        search(child);
      }
      return;
    }
    case 8 /* COMPOUND_EXPRESSION */: {
      if (!node.ast) {
        return;
      }
      addExpression({ loc: node.loc, src: node.loc.source });
      return;
    }
    // case NodeTypes.IF:
    // case NodeTypes.FOR:
    // case NodeTypes.TEXT_CALL:
    default: {
      throw new Error(`Unexpected node type: ${node.type}`);
    }
  }
}
async function transpileVueTemplate(content, root, offset = 0, transform) {
  const { MagicString } = await import('vue/compiler-sfc');
  const expressions = [];
  const s = new MagicString(content);
  handleNode(root, (...items) => expressions.push(...items));
  await Promise.all(
    expressions.map(async (item) => {
      if (item.src.trim() === "") {
        item.replacement = item.src;
        return;
      }
      try {
        item.replacement = await transformJsSnippet(item.src, transform);
        const surrounding = getSurrounding(
          content,
          item.loc.start.offset - offset,
          item.loc.end.offset - offset
        );
        if (surrounding) {
          const replace = surrounding.code === `"` ? `'` : `"`;
          item.replacement = replaceQuote(
            item.replacement,
            surrounding.code,
            replace
          );
        }
      } catch {
        item.replacement = item.src;
      }
    })
  );
  for (const item of expressions) {
    if (item.replacement) {
      s.overwrite(
        item.loc.start.offset - offset,
        item.loc.end.offset - offset,
        item.replacement
      );
    }
  }
  return s.toString();
}
function replaceQuote(code, target, replace) {
  let res = code;
  if (res.includes(target)) {
    res = res.replaceAll(replace, `\\${replace}`);
    res = res.replaceAll(target, replace);
  }
  return res;
}
function getSurrounding(code, start, end) {
  const empty = /* @__PURE__ */ new Set([" ", "\n", "\r", "	"]);
  let startIndex = start - 1;
  let endIndex = end;
  while (startIndex > 0 && empty.has(code.at(startIndex))) {
    startIndex--;
  }
  while (endIndex < code.length && empty.has(code.at(endIndex))) {
    endIndex++;
  }
  const prev = startIndex >= 0 ? code.at(startIndex) : "";
  const next = endIndex < code.length ? code.at(endIndex) : "";
  return prev && next && prev === next ? { code: prev, prevAt: startIndex, nextAt: endIndex } : void 0;
}
async function transformJsSnippet(code, transform) {
  const isObject = /^\s*\{.*\}\s*$/.test(code);
  let res = isObject ? await transform(code) : await transform(`(${code})`);
  res = res.trim();
  if (res.endsWith(";")) {
    res = res.slice(0, -1);
  }
  return res;
}

export { preTranspileScriptSetup as p, transpileVueTemplate as t };
